@@ -1,5 +1,5 @@
 #
-# Copyright (C) 1997, 1998, 1999  Motoyuki Kasahara
+# Copyright (C) 1997, 98, 99, 2000  Motoyuki Kasahara
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -22,12 +22,14 @@ set ndtp_socket {}
 set ndtp_books {0}
 
 # title list of the server
-set ndtp_titles(0) "(not selected)"
+set ndtp_titles(0) [_ "(not selected)"]
 
 # page information for each book
 set ndtp_pages_menu(0) 0
 set ndtp_pages_copyright(0) 0
 set ndtp_pages_word(0) 0
+set ndtp_pages_endword(0) 0
+set ndtp_pages_keyword(0) 0
 
 # bitmap size information for each book.
 set ndtp_sizes_16(0) 0
@@ -37,13 +39,13 @@ set ndtp_sizes_48(0) 0
 
 # Do `ndtp_pages_menu', `menu_copyright' and `ndtp_pages_word' in
 # a book have valid values?
-set ndtp_cache_pages(0) 0
+set ndtp_cache_indexes(0) 0
 
 # Do `size_16' ... `size_48' for a book have valid values?
 set ndtp_cache_sizes(0) 0
 
 # Do bitmaps are created for a book?
-set ndtp_cache_created(0) 0
+set ndtp_cache_all_bitmaps(0) 0
 
 # current book
 set ndtp_current_book 0
@@ -52,11 +54,14 @@ set ndtp_current_book 0
 set ndtp_current_size 0
 
 # created bitmap size
-set ndtp_created_size 0
+set ndtp_cache_size 0
 
 # whether server accept bitmap extension command?
 set ndtp_extension_ok {}
 
+# default method
+set ndtp_default_method "exactword"
+
 #
 # Examine whether a connection is active.
 #
@@ -78,18 +83,20 @@ proc ndtp_initialize {} {
     global ndtp_books
     global ndtp_current_book
     global ndtp_current_size
-    global ndtp_created_size
+    global ndtp_cache_size
     global ndtp_titles
     global ndtp_pages_menu
     global ndtp_pages_copyright
     global ndtp_pages_word
+    global ndtp_pages_endword
+    global ndtp_pages_keyword
     global ndtp_sizes_16
     global ndtp_sizes_24
     global ndtp_sizes_30
     global ndtp_sizes_48
-    global ndtp_cache_pages
+    global ndtp_cache_indexes
     global ndtp_cache_sizes
-    global ndtp_cache_created
+    global ndtp_cache_all_bitmaps
     global ndtp_extension_ok
 
     #
@@ -112,16 +119,18 @@ proc ndtp_initialize {} {
 	catch {unset ndtp_pages_menu($i)}
 	catch {unset ndtp_pages_copyright($i)}
 	catch {unset ndtp_pages_word($i)}
+	catch {unset ndtp_pages_endword($i)}
+	catch {unset ndtp_pages_keyword($i)}
 	catch {unset ndtp_sizes_16($i)}
 	catch {unset ndtp_sizes_24($i)}
 	catch {unset ndtp_sizes_30($i)}
 	catch {unset ndtp_sizes_48($i)}
-	catch {unset ndtp_cache_pages($i)}
+	catch {unset ndtp_cache_indexes($i)}
 	catch {unset ndtp_cache_sizes($i)}
-	catch {unset ndtp_cache_created($i)}
+	catch {unset ndtp_cache_all_bitmaps($i)}
     }
 
-    set ndtp_created_size 0
+    set ndtp_cache_size 0
 
     set ndtp_books {}
     set ndtp_extension_ok {}
@@ -138,9 +147,9 @@ proc ndtp_open {hostname port books} {
     global ndtp_socket
     global ndtp_books
     global ndtp_titles
-    global ndtp_cache_pages
+    global ndtp_cache_indexes
     global ndtp_cache_sizes
-    global ndtp_cache_created
+    global ndtp_cache_all_bitmaps
 
     upvar $books up_books
     set up_books {}
@@ -186,7 +195,7 @@ proc ndtp_open {hostname port books} {
 	    ndtp_close
 	}
 	ndtp_initialize
-	set ndtp_error_message "Authentication failed."
+	set ndtp_error_message [_ "Authentication failed."]
 	return 0
     }
 
@@ -203,14 +212,14 @@ proc ndtp_open {hostname port books} {
 	    $line dummy dummy title bookcode]} {
 	lappend ndtp_books $bookcode
 	set ndtp_titles($bookcode) $title
-	set ndtp_cache_pages($bookcode) 0
+	set ndtp_cache_indexes($bookcode) 0
 	set ndtp_cache_sizes($bookcode) 0
-	set ndtp_cache_created($bookcode) 0
+	set ndtp_cache_all_bitmaps($bookcode) 0
     } elseif {$line == "\$*"} {
         return 1
     } elseif {$line != "\$*"} {
 	ndtp_initialize
-	set ndtp_error_message "Cannot get a book list."
+	set ndtp_error_message [_ "Cannot get a book list."]
 	return 0
     }
     while {1} {
@@ -223,14 +232,14 @@ proc ndtp_open {hostname port books} {
 	}
 	lappend ndtp_books $bookcode
 	set ndtp_titles($bookcode) $title
-	set ndtp_cache_pages($bookcode) 0
+	set ndtp_cache_indexes($bookcode) 0
 	set ndtp_cache_sizes($bookcode) 0
-	set ndtp_cache_created($bookcode) 0
+	set ndtp_cache_all_bitmaps($bookcode) 0
     }
     if {$line != "\$*"} {
 	puts $line
 	ndtp_initialize
-	set ndtp_error_message "Cannot get a book list."
+	set ndtp_error_message [_ "Cannot get a book list."]
 	return 0
     }
 
@@ -275,16 +284,21 @@ proc ndtp_select_book {book} {
     global ndtp_pages_menu
     global ndtp_pages_copyright
     global ndtp_pages_word
+    global ndtp_pages_endword
+    global ndtp_pages_keyword
     global ndtp_sizes_16
     global ndtp_sizes_24
     global ndtp_sizes_30
     global ndtp_sizes_48
-    global ndtp_cache_pages
+    global ndtp_cache_indexes
     global ndtp_cache_sizes
     global ndtp_extension_ok
 
-    if {$book == "nodict"} {
+    if {$book == "nodict" || $book == "0"} {
 	set book 0
+    } elseif {[lsearch -exact $ndtp_books $book] == -1} {
+	set ndtp_error_message [format [_ "Unknown book %s."] $book]
+	return 0
     }
 
     #
@@ -298,10 +312,12 @@ proc ndtp_select_book {book} {
     # If this book has not selected yet, initialize page and
     # bitmap size tables.
     #
-    if {!$ndtp_cache_pages($book)} {
+    if {!$ndtp_cache_indexes($book)} {
 	set ndtp_pages_menu($book) 0
 	set ndtp_pages_copyright($book) 0
 	set ndtp_pages_word($book) 0
+	set ndtp_pages_endword($book) 0
+	set ndtp_pages_keyword($book) 0
     }
     if {!$ndtp_cache_sizes($book)} {
 	set ndtp_sizes_16($book) 0
@@ -316,12 +332,12 @@ proc ndtp_select_book {book} {
     set ndtp_current_book 0
     set ndtp_current_size 0
     if {![ndtp_put_line "L$book"]} {
-	set ndtp_cache_pages($book) 0
+	set ndtp_cache_indexes($book) 0
 	set ndtp_cache_sizes($book) 0
 	return 0
     }
     if {![ndtp_get_line line]} {
-	set ndtp_cache_pages($book) 0
+	set ndtp_cache_indexes($book) 0
 	set ndtp_cache_sizes($book) 0
 	return 0
     }
@@ -329,8 +345,8 @@ proc ndtp_select_book {book} {
 	if {![string match {$?} $line]} {
 	    ndtp_initialize
 	}
-	set ndtp_error_message "Cannot select the book."
-	set ndtp_cache_pages($book) 0
+	set ndtp_error_message [format [_ "Cannot select the book %s."] $book]
+	set ndtp_cache_indexes($book) 0
 	set ndtp_cache_sizes($book) 0
 	return 0
     }
@@ -347,7 +363,7 @@ proc ndtp_select_book {book} {
     # Send the `I' command.
     # (If fails, give up talking to the server.)
     #
-    if {!$ndtp_cache_pages($ndtp_current_book)} {
+    if {!$ndtp_cache_indexes($ndtp_current_book)} {
 	if {![ndtp_put_line "I"]} {
 	    return 0
 	}
@@ -356,7 +372,8 @@ proc ndtp_select_book {book} {
 	}
 	if {$line != "\$I"} {
 	    ndtp_initialize
-	    set ndtp_error_message "Cannot get a method list."
+	    set ndtp_error_message [format \
+		    [_ "Cannot get a method list of %s."] $book]
 	    return 0
 	}
 
@@ -369,16 +386,19 @@ proc ndtp_select_book {book} {
 	    }
 	    switch -exact $method {
 		"IA" {set ndtp_pages_word($ndtp_current_book) $page}
+		"BA" {set ndtp_pages_endword($ndtp_current_book) $page}
+		"JO" {set ndtp_pages_keyword($ndtp_current_book) $page}
 		"HA" {set ndtp_pages_menu($ndtp_current_book) $page}
 		"OK" {set ndtp_pages_copyright($ndtp_current_book) $page}
 	    }
 	}
 	if {$line != "\$\$"} {
 	    ndtp_initialize
-	    set ndtp_error_message "Cannot get a method list."
+	    set ndtp_error_message [format \
+		    [_ "Cannot get a method list of %s."] $book]
 	    return 0
 	}
-	set ndtp_cache_pages($ndtp_current_book) 1
+	set ndtp_cache_indexes($ndtp_current_book) 1
     }
 
     #
@@ -393,16 +413,20 @@ proc ndtp_select_book {book} {
 	if {![ndtp_get_line line]} {
 	    return 0
 	}
-	if {$line != "\$?"} {
+	if {$line == "\$?"} {
+	    set ndtp_extension_ok 0
+	} else {
 	    if {$line != "\$I"} {
 		if {[string match {$?} $line]} {
 		    ndtp_select_book 0
 		} else {
 		    ndtp_initialize
 		}
-		set ndtp_error_message "Cannot get a bitmap size list."
+		set ndtp_error_message [format \
+			[_ "Cannot get a bitmap size list of %s."] $book]
 		return 0
 	    }
+	    set ndtp_extension_ok 1
 	    while {1} {
 		if {![ndtp_get_line line]} {
 		    return 0
@@ -419,7 +443,8 @@ proc ndtp_select_book {book} {
 	    }
 	    if {$line != "\$\$"} {
 		ndtp_initialize
-		set ndtp_error_message "Cannot get a bitmap size list."
+		set ndtp_error_message [format \
+			[_ "Cannot get a bitmap size list of %s."] $book]
 		return 0
 	    }
 	}
@@ -434,6 +459,7 @@ proc ndtp_select_book {book} {
 #
 proc ndtp_query_word {word positions headings} {
     global ndtp_error_message
+    global ndtp_default_method
     upvar $positions up_positions
     upvar $headings up_headings
 
@@ -445,35 +471,47 @@ proc ndtp_query_word {word positions headings} {
     # Delete whitespaces in the beginning and tail of the word.
     #
     regsub -all "\n" $word "" word
-    regsub {^[ 	]*(.*)[ 	]*$} $word {\1} word
+    regsub -all "\t" $word " " word
+    regsub {^ *(.*) *$} $word {\1} word
 
     #
-    # Determine an index type.
+    # Adjust $word according with $ndtp_default_method.
     #
-    if {[kanji code $word] != {ANY}} {
-        if {[kanji string match {\**} $word]} {
-	    set length [kstring length $word]
-	    set revword ""
-	    for {set i 0} {$i < $length} {incr i} {
-		set c [kstring index $word $i]
-		set revword "$c$revword"
-	    }
-	    set line "PK$revword"
-        } else {
-	    set line "Pk$word"
-        }
+    if {$ndtp_default_method == "keyword" \
+	    && ![kanji string match {*;*} $word]} {
+	set word "$word;"
+    } elseif {$ndtp_default_method == "endword" \
+	    && ![kanji string match {\**} $word]} {
+	set word "*$word"
+    } elseif {$ndtp_default_method == "word" \
+	    && ![kanji string match {*\*} $word]} {
+	set word "$word*"
+    }
+
+    #
+    # Set a request line.
+    #
+    if {[kanji string match {*;*} $word]} {
+	regsub -all {[ 	]*;[ 	]*} $word {	} word
+	set line "Pj$word"
+    } elseif {[kanji string match {\**} $word]} {
+	set length [kstring length $word]
+	set reversed_word ""
+	for {set i 0} {$i < $length} {incr i} {
+	    set c [kstring index $word $i]
+	    set reversed_word "$c$reversed_word"
+	}
+	if {[kanji code $word] == {ANY}} {
+	    set line "PA$reversed_word"
+	} else {
+	    set line "PK$reversed_word"
+	}
     } else {
-        if {[string match {\**} $word]} {
-	    set length [string length $word]
-	    set revword ""
-	    for {set i 0} {$i < $length} {incr i} {
-		set c [string index $word $i]
-		set revword "$c$revword"
-	    }
-	    set line "PA$revword"
-        } else {
-            set line "Pa$word"
-        }
+	if {[kanji code $word] == {ANY}} {
+	    set line "Pa$word"
+	} else {
+	    set line "Pk$word"
+	}
     }
 
     #
@@ -489,7 +527,7 @@ proc ndtp_query_word {word positions headings} {
 	if {![string match {$?} $line]} {
 	    ndtp_initialize
 	}
-	set ndtp_error_message "Cannot search the word"
+	set ndtp_error_message [_ "Cannot search the word."]
 	return 0
     }
     while {1} {
@@ -527,14 +565,14 @@ proc ndtp_get_text {position text} {
     switch -exact $position {
 	"menu" {
 	    if {$ndtp_pages_menu($ndtp_current_book) == 0} {
-		set ndtp_error_message "This book doesn't have a menu"
+		set ndtp_error_message [_ "This book doesn't have a menu"]
 		return 0
 	    }
 	    set position "$ndtp_pages_menu($ndtp_current_book):0"
 	}
 	"copyright" {
 	    if {$ndtp_pages_copyright($ndtp_current_book) == 0} {
-		set ndtp_error_message "This book doesn't have a menu"
+		set ndtp_error_message [_ "This book doesn't have a menu"]
 		return 0
 	    }
 	    set position "$ndtp_pages_copyright($ndtp_current_book):0"
@@ -554,7 +592,7 @@ proc ndtp_get_text {position text} {
 	if {![string match {$?} $line]} {
 	    ndtp_initialize
 	}
-	set ndtp_error_message "Cannot get text."
+	set ndtp_error_message [_ "Cannot get text."]
 	return 0
     }
 
@@ -575,7 +613,7 @@ proc ndtp_get_text {position text} {
 #
 # Select a bitmap size, and get all bitmap data.
 #
-proc ndtp_select_bitmap_size {sizes} {
+proc ndtp_select_bitmap_size {size {load_mode "pre-load"}} {
     global ndtp_error_message
     global ndtp_sizes_16
     global ndtp_sizes_24
@@ -583,27 +621,8 @@ proc ndtp_select_bitmap_size {sizes} {
     global ndtp_sizes_48
     global ndtp_current_book
     global ndtp_current_size
-    global ndtp_created_size
-    global ndtp_cache_created
-
-    #
-    # Determine bitmap size.
-    #
-    set size 0
-    foreach i $sizes {
-	if {$i == "16" && $ndtp_sizes_16($ndtp_current_book)} {
-	    set size 16
-	} elseif {$i == "24" && $ndtp_sizes_24($ndtp_current_book)} {
-	    set size 24
-	} elseif {$i == "30" && $ndtp_sizes_30($ndtp_current_book)} {
-	    set size 30
-	} elseif {$i == "48" && $ndtp_sizes_48($ndtp_current_book)} {
-	    set size 48
-	}
-	if {$size != 0} {
-	    break
-	}
-    }
+    global ndtp_cache_size
+    global ndtp_cache_all_bitmaps
 
     #
     # If $book has been selected, return immediately.
@@ -626,7 +645,7 @@ proc ndtp_select_bitmap_size {sizes} {
 	if {$line != "\$<" && $line != "\$&"} {
 	    ndtp_initialize
 	}
-	set ndtp_error_message "Cannot select the font."
+	set ndtp_error_message [_ "Cannot select the font."]
 	return 0
     }
     set ndtp_current_size $size
@@ -641,7 +660,7 @@ proc ndtp_select_bitmap_size {sizes} {
     #
     # Delete bitmaps if bitmas of different size have been created.
     #
-    if {$size != $ndtp_created_size} {
+    if {$size != $ndtp_cache_size} {
 	ndtp_delete_bitmaps
     }
 
@@ -650,7 +669,9 @@ proc ndtp_select_bitmap_size {sizes} {
     # Return with a success code even when the server responds with
     # the error code `unknown command'.
     #
-    if {!$ndtp_cache_created($ndtp_current_book)} {
+    if {$load_mode != "pre-load"} {
+	set ndtp_cache_size $size
+    } elseif {!$ndtp_cache_all_bitmaps($ndtp_current_book)} {
 	if {![ndtp_put_line "XB"]} {
 	    return 0
 	}
@@ -663,7 +684,7 @@ proc ndtp_select_bitmap_size {sizes} {
 	    } else {
 		ndtp_select_bitmap_size {}
 	    }
-	    set ndtp_error_message "Cannot get a font list."
+	    set ndtp_error_message [_ "Cannot get font images."]
 	    return 0
 	}
 
@@ -689,18 +710,81 @@ proc ndtp_select_bitmap_size {sizes} {
 	if {$imageid != {}} {
 	    image create bitmap "$ndtp_current_book/$imageid" -data $xbmdata
 	}
-	set ndtp_created_size $size
-	set ndtp_cache_created($ndtp_current_book) 1
+	set ndtp_cache_size $size
+	set ndtp_cache_all_bitmaps($ndtp_current_book) 1
     }
 
     return 1
 }
 
+#
+# Get bitmap data of character `imageid', if there is no cache data.
+#
+proc ndtp_get_bitmap {imageid} {
+    global ndtp_error_message
+    global ndtp_current_book
+    global ndtp_current_size
+
+    #
+    # Return immediately if bitmap size is not selected.
+    #
+    if {$ndtp_current_size == 0} {
+	return 0
+    }
+
+    #
+    # Return immediately if we have already got the cache data.
+    #
+    if {![catch {image type "$ndtp_current_book/$imageid"}]} {
+	return 1
+    }
+
+    #
+    # Send the `Xb' command.
+    #
+    if {![ndtp_put_line "Xb$imageid"]} {
+	return 0
+    }
+    if {![ndtp_get_line line]} {
+	return 0
+    }
+    if {$line != "\$I"} {
+	if {$line != "\$<" && $line != "\$&" && $line != "\$?"} {
+	    ndtp_initialize
+	} else {
+	    ndtp_select_bitmap_size {}
+	}
+	set ndtp_error_message [_ "Cannot get a font image."]
+	return 0
+    }
+
+    set xbmdata {}
+    while {1} {
+	if {![ndtp_get_line line]} {
+	    return 0
+	}
+	if {$line == "\$\$"} {
+	    break
+	} else {
+	    append xbmdata $line "\n"
+	}
+    }
+
+    if {$xbmdata != {}} {
+	image create bitmap "$ndtp_current_book/$imageid" -data $xbmdata
+	return 1
+    } else {
+	return 0
+    }
+
+    # not reached
+}
+
 #
 # Delete images.
 #
 proc ndtp_delete_bitmaps {} {
-    global ndtp_created_size
+    global ndtp_cache_size
 
     foreach i [image names] {
 	if {[string match "*/*" $i]} {
@@ -708,16 +792,34 @@ proc ndtp_delete_bitmaps {} {
 	}
     }
 
-    set ndtp_created_size 0
+    set ndtp_cache_size 0
 }
 
 #
 # The title of $book.
 #
 proc ndtp_title {book} {
+    global ndtp_books
     global ndtp_titles
 
-    return $ndtp_titles($book)
+    if {$book == 0 || $book == "nodict"} {
+	return [_ "(not selected)"]
+    } elseif {[lsearch -exact $ndtp_books $book] == -1} {
+	return ""
+    } else {
+	return $ndtp_titles($book)
+    }
+
+    # not reached
+}
+
+#
+# A list of books.
+#
+proc ndtp_books {} {
+    global ndtp_books
+
+    return $ndtp_books
 }
 
 #
@@ -733,6 +835,32 @@ proc ndtp_have_word_search {book} {
     }
 }
 
+#
+# Does the current book have endword search?
+#
+proc ndtp_have_endword_search {book} {
+    global ndtp_pages_endword
+
+    if {$ndtp_pages_endword($book) != 0} {
+	return 1
+    } else {
+	return 0
+    }
+}
+
+#
+# Does the current book have endword search?
+#
+proc ndtp_have_keyword_search {book} {
+    global ndtp_pages_keyword
+
+    if {$ndtp_pages_keyword($book) != 0} {
+	return 1
+    } else {
+	return 0
+    }
+}
+
 #
 # Does the current book have menu?
 #
@@ -773,7 +901,7 @@ proc ndtp_get_line {line} {
 	return 0
     } elseif {[eof $ndtp_socket]} {
 	ndtp_initialize
-	set ndtp_error_message "The server closes the connection."
+	set ndtp_error_message [_ "The server closes the connection."]
 	return 0
     }
 
@@ -808,3 +936,17 @@ proc ndtp_put_line {line} {
     return 1
 }
 
+#
+# Set default method.
+#
+proc ndtp_set_default_method {method} {
+    global ndtp_default_method
+
+    if {$method == "exactword" \
+	    || $method == "word" \
+	    || $method == "endword" \
+	    || $method == "keyword"} {
+	set ndtp_default_method $method
+    }
+}
+
